from chainer import cuda, optimizers, Variable, serializers
import numpy as np
from PIL import Image
import os
import datetime

from model import*


epoch_num = 20
batch_size = 15
chanel_num = 3  
image_file_num = 6
block_num = 10
block_set_num = image_file_num*block_num
block_size = 64
train_block_size = 60
light_kind_num = 6
motion_num = 25
train_data_num = chanel_num*block_set_num*light_kind_num*motion_num
check_num = train_data_num
check_num = check_num - check_num%15

shots_num = 3
view_point_row_num = 5
view_point_num = view_point_row_num*view_point_row_num
dev = 0.005

aug_data = np.zeros((batch_size, view_point_num, block_size, block_size), np.float32)
aug_data_frame1 = np.zeros((batch_size, view_point_num, train_block_size, train_block_size), np.float32)
aug_data_frame2 = np.zeros((batch_size, view_point_num, train_block_size, train_block_size), np.float32)
aug_data_frame3 = np.zeros((batch_size, view_point_num, train_block_size, train_block_size), np.float32)

cnt = 0
loss_sum = 0.0

xp = cuda.cupy
cuda.get_device(0).use()


if os.path.isdir("./Dst") == False:
	os.mkdir("./Dst")

if os.path.isdir("./Dst/Mask") == False:
	os.mkdir("./Dst/Mask")

if os.path.isdir("./Dst/Model") == False:
	os.mkdir("./Dst/Model")




#Load train datas
load_light_field = np.zeros((block_set_num*chanel_num, view_point_num, block_size, block_size), np.uint8)
start_index_R = 0;
start_index_G = image_file_num*block_num
start_index_B = image_file_num*block_num*2

print("train with %f devietion noise" %dev)
for n in range(image_file_num):
	if(n % 10 == 0):
		print("load %04d" % (n))

	fname = "./TrainingData/%03d.png" % (n)
	img_load = Image.open(fname)
	img_load = np.asarray(img_load)

	for b in range(block_num):
		for t in range(view_point_num):
			load_light_field[n*block_num+b +start_index_R,t,:,:] = img_load[block_size*b:block_size*b+block_size,block_size*t:block_size*t+block_size,0]
			load_light_field[n*block_num+b +start_index_G,t,:,:] = img_load[block_size*b:block_size*b+block_size,block_size*t:block_size*t+block_size,1]
			load_light_field[n*block_num+b +start_index_B,t,:,:] = img_load[block_size*b:block_size*b+block_size,block_size*t:block_size*t+block_size,2]





#Model initaition
shot_A = Shot()
shot_A.to_gpu()
shot_A_optimizer = optimizers.Adam(alpha=0.0001, beta1=0.9, beta2=0.999, eps=1e-08)
shot_A_optimizer.setup(shot_A)

shot_B = Shot()
shot_B.to_gpu()
shot_B_optimizer = optimizers.Adam(alpha=0.0001, beta1=0.9, beta2=0.999, eps=1e-08)
shot_B_optimizer.setup(shot_B)

reconstruct = Reconstruct()
reconstruct.to_gpu()
reconstruct_optimizer = optimizers.Adam(alpha=0.0001, beta1=0.9, beta2=0.999, eps=1e-08)
reconstruct_optimizer.setup(reconstruct)

vdsr = VeryDeepSuperResolution()
vdsr.to_gpu()
vdsr_optimizer = optimizers.Adam(alpha=0.0001, beta1=0.9, beta2=0.999, eps=1e-08)
vdsr_optimizer.setup(vdsr)


#Physical limitations of masks
shot_A.cn01.W.data[shot_A.cn01.W.data < 0.0] = 0.0
shot_A.cn01.W.data[shot_A.cn01.W.data > 1.0] = 1.0
shot_A.cn01.b.data[:] = 0.0
shot_B.cn01.W.data[shot_B.cn01.W.data < 0.0] = 0.0
shot_B.cn01.W.data[shot_B.cn01.W.data > 1.0] = 1.0
shot_B.cn01.b.data[:] = 0.0



#Train loop
for i in range(1, epoch_num+1):

	sff = np.random.permutation(train_data_num)

	for n in range(0, train_data_num, batch_size):
		#Data augmentation by intensity levels
		for j in range(batch_size):
			aug_data[j,:,:,:] = ((1-(int(sff[n+j]/motion_num)%light_kind_num)/10)*load_light_field[int(sff[n+j]/light_kind_num/motion_num)]/255).astype(np.float32)

                        #Creating pseudo motion
			shift_x = int((sff[n+j]%motion_num)%5) -2
			shift_y = int((sff[n+j]%motion_num)/5) -2
			aug_data_frame1[j,:,:,:]= aug_data[j, : , (2-shift_x):(2+train_block_size-shift_x), (2-shift_y):(2+train_block_size-shift_y)]
			aug_data_frame2[j,:,:,:]= aug_data[j, : , 2:2+train_block_size, 2:2+train_block_size]
			aug_data_frame3[j,:,:,:]= aug_data[j, : , (2+shift_x):(2+train_block_size+shift_x), (2+shift_y):(2+train_block_size+shift_y)]


		gray_light_field_frame1 = Variable(cuda.to_gpu(aug_data_frame1))
		gray_light_field_frame2 = Variable(cuda.to_gpu(aug_data_frame2))
		gray_light_field_frame3 = Variable(cuda.to_gpu(aug_data_frame3))


		#Train models		
		shot_A.cleargrads()
		shot_B.cleargrads()
		reconstruct.cleargrads()
		vdsr.cleargrads()


		gray_acquired_image1A = shot_A(gray_light_field_frame1)
		gray_acquired_image2B = shot_B(gray_light_field_frame2)
		gray_acquired_image3A = shot_A(gray_light_field_frame3)

		gray_acquired_images_ABA = F.concat((gray_acquired_image1A, gray_acquired_image2B, gray_acquired_image3A), axis=1)
		gray_noise_acquired_images_ABA = gray_acquired_images_ABA + view_point_num*dev*xp.random.randn(batch_size,shots_num,train_block_size,train_block_size)

		gray_makeshift_light_field = reconstruct(gray_noise_acquired_images_ABA)
		gray_reconstruct_light_field = vdsr(gray_makeshift_light_field)

		loss = F.mean_squared_error(gray_reconstruct_light_field, gray_light_field_frame2)

		loss.backward()
		shot_A_optimizer.update()
		shot_B_optimizer.update()
		reconstruct_optimizer.update()
		vdsr_optimizer.update()

		shot_A.cn01.W.data[shot_A.cn01.W.data < 0.0] = 0.0
		shot_A.cn01.W.data[shot_A.cn01.W.data > 1.0] = 1.0
		shot_A.cn01.b.data[:] = 0.0
		shot_B.cn01.W.data[shot_B.cn01.W.data < 0.0] = 0.0
		shot_B.cn01.W.data[shot_B.cn01.W.data > 1.0] = 1.0
		shot_B.cn01.b.data[:] = 0.0

		print("epoch:%04d/%04d, data:%06d/%06d, psnr:%.6f" % (i, epoch_num, n, train_data_num, 10*xp.log10(1.0/loss.data)))

		loss_sum += loss.data	
		cnt += 1

		#Save masks and calculate average psnr
		if(n%check_num == 0):

			mask_txt_name = "./Dst/Mask/mask_txt_%03depoch-%06ddata.txt" % (i, n)
			f_mask = open(mask_txt_name, "w")

			img_save = cuda.to_cpu(shot_A.cn01.W.data[0,:,0,0])
			img_save = img_save.reshape(view_point_row_num,view_point_row_num)
			f_mask.write("%s\n" % img_save)	
			img_save = Image.fromarray(np.uint8(img_save * 255.0))
			img_save = img_save.resize((256,256))
			img_save.save("./Dst/Mask/Mask_A_epoch{:04d}-batch{:06d}done.png".format(i, n))

			img_save = cuda.to_cpu(shot_B.cn01.W.data[0,:,0,0])
			img_save = img_save.reshape(view_point_row_num,view_point_row_num)
			f_mask.write("%s\n" % img_save)	
			img_save = Image.fromarray(np.uint8(img_save * 255.0))
			img_save = img_save.resize((256,256))
			img_save.save("./Dst/Mask/Mask_B_epoch{:04d}-batch{:06d}done.png".format(i, n))

			f_mask.close()


			print("@@@    average psnr:%.6f    @@@" % (10*xp.log10(1.0 / (loss_sum / cnt))))
			f_psnr = open("./Dst/average_psnr.txt", "a")
			f_psnr.write("epoch:%04d/%04d, data:%06d/%06d, psnr:%.6f\n" % (i, epoch_num, n, train_data_num, 10*xp.log10(1.0 / (loss_sum / cnt))))
			f_psnr.close()

			cnt = 0
			loss_sum = 0.0
	


	#Save models every epoch
	shot_A.to_cpu()
	serializers.save_npz("./Dst/Model/shot_A%04d.npz" %i, shot_A)
	shot_A.to_gpu()

	shot_B.to_cpu()
	serializers.save_npz("./Dst/Model/shot_B%04d.npz" %i, shot_B)
	shot_B.to_gpu()

	reconstruct.to_cpu()
	serializers.save_npz("./Dst/Model/reconstruct%04d.npz" %i, reconstruct)
	reconstruct.to_gpu()

	vdsr.to_cpu()
	serializers.save_npz("./Dst/Model/vdsr%04d.npz" %i, vdsr)
	vdsr.to_gpu()














